from pyTextMiner.collector.colBasic import *
from pyTextMiner.collector.colNaverNews import *
from pyTextMiner.collector.colChosun import *
from pyTextMiner.collector.colDonga import *
from pyTextMiner.collector.colJoongang import *
from pyTextMiner.collector.colBigkinds import *
from pyTextMiner.collector.colNaverBlog import *